from __future__ import absolute_import, print_function

import logging
import re
import time
from functools import cached_property
from typing import Any, Dict, List, Optional, Tuple, Union

import adbutils

from uiautomator2.core import BasicUiautomatorServer
from uiautomator2._proto import HTTP_TIMEOUT, SCROLL_STEPS, Direction
from uiautomator2.exceptions import *
from uiautomator2.settings import Settings
from uiautomator2.utils import image_convert, list2cmdline, deprecated
from uiautomator2.abstract import ShellResponse

logger = logging.getLogger(__name__)


class _BaseClient(BasicUiautomatorServer):
    """
    提供最基础的控制类，这个类暂时先不公开吧
    """

    def __init__(self, serial: Optional[Union[str, adbutils.AdbDevice]] = None):
        """
        Args:
            serial: device serialno
        """
        if isinstance(serial, adbutils.AdbDevice):
            self.__serial = serial.serial
            self._dev = serial
        else:
            self.__serial = serial
            self._dev = self._wait_for_device()
        self._debug = False
        BasicUiautomatorServer.__init__(self, self._dev)
    
    @property
    def _serial(self) -> str:
        return self.__serial
    
    def _wait_for_device(self, timeout=10) -> adbutils.AdbDevice:
        """
        wait for device came online, if device is remote, reconnect every 1s

        Returns:
            adbutils.AdbDevice
        
        Raises:
            ConnectError
        """
        for d in adbutils.adb.device_list():
            if d.serial == self._serial:
                return d

        _RE_remote_adb = re.compile(r"^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}:\d+$")
        _is_remote = _RE_remote_adb.match(self._serial) is not None

        adb = adbutils.adb
        deadline = time.time() + timeout
        while time.time() < deadline:
            title = "device reconnecting" if _is_remote else "wait-for-device"
            logger.debug("%s, time left(%.1fs)", title, deadline - time.time())
            if _is_remote:
                try:
                    adb.disconnect(self._serial)
                    adb.connect(self._serial, timeout=1)
                except (adbutils.AdbError, adbutils.AdbTimeout) as e:
                    logger.debug("adb reconnect error: %s", str(e))
                    time.sleep(1.0)
                    continue
            try:
                adb.wait_for(self._serial, timeout=1)
            except (adbutils.AdbError, adbutils.AdbTimeout):
                continue
            return adb.device(self._serial)
        raise ConnectError(f"device {self._serial} not online")

    @property
    def adb_device(self) -> adbutils.AdbDevice:
        return self._dev
    
    @cached_property
    def settings(self) -> Settings:
        return Settings(self)

    def sleep(self, seconds: float):
        """ same as time.sleep """
        time.sleep(seconds)

    def shell(self, cmdargs: Union[str, List[str]], timeout=60) -> ShellResponse:
        """
        Run shell command on device

        Args:
            cmdargs: str or list, example: "ls -l" or ["ls", "-l"]
            timeout: seconds of command run, works on when stream is False

        Returns:
            ShellResponse

        Raises:
            AdbShellError
        """
        try:
            if self.debug:
                print("shell:", list2cmdline(cmdargs))
            logger.debug("shell: %s", list2cmdline(cmdargs))
            ret = self._dev.shell2(cmdargs, timeout=timeout)
            return ShellResponse(ret.output, ret.returncode)
        except adbutils.AdbError as e:
            raise AdbShellError(e)

    @property
    def info(self) -> Dict[str, Any]:
        return self.jsonrpc.deviceInfo(http_timeout=10)
    
    @property
    def device_info(self) -> Dict[str, Any]:
        serial = self._dev.getprop("ro.serialno")
        sdk = self._dev.getprop("ro.build.version.sdk")
        version = self._dev.getprop("ro.build.version.release")
        brand = self._dev.getprop("ro.product.brand")
        model = self._dev.getprop("ro.product.model")
        arch = self._dev.getprop("ro.product.cpu.abi")
        return {
            "serial": serial,
            "sdk": int(sdk) if sdk.isdigit() else None,
            "brand": brand,
            "model": model,
            "arch": arch,
            "version": int(version) if version.isdigit() else None,
        }

    @property
    def wlan_ip(self) -> Optional[str]:
        try:
            return self._dev.wlan_ip()
        except adbutils.AdbError:
            return None

    @property
    def jsonrpc(self):
        class JSONRpcWrapper():
            def __init__(self, server: BasicUiautomatorServer):
                self.server = server
                self.method = None

            def __getattr__(self, method):
                self.method = method  # jsonrpc function name
                return self

            def __call__(self, *args, **kwargs):
                http_timeout = kwargs.pop('http_timeout', HTTP_TIMEOUT)
                params = args if args else kwargs
                return self.server.jsonrpc_call(self.method, params, http_timeout)

        return JSONRpcWrapper(self)

    def reset_uiautomator(self):
        """
        restart uiautomator service

        Orders:
            - stop uiautomator keeper
            - am force-stop com.github.uiautomator
            - start uiautomator keeper(am instrument -w ...)
            - wait until uiautomator service is ready
        """
        self.stop_uiautomator()
        self.start_uiautomator()

    def push(self, src, dst: str, mode=0o644):
        """
        Push file into device

        Args:
            src (path or fileobj): source file
            dst (str): destination can be folder or file path
            mode (int): file mode
        """
        self._dev.sync.push(src, dst, mode=mode)

    def pull(self, src: str, dst: str):
        """
        Pull file from device to local
        """
        try:
            self._dev.sync.pull(src, dst, exist_ok=True)
        except TypeError:
            self._dev.sync.pull(src, dst)



# coding: utf-8
#
# BaseException
#   +- RPCError
#   |   +- RPCUnknownError
#   |   +- RPCInvalidError
#   |   +- HierarchyEmptyError
#   |   +- RPCStackOverflowError
#   |   +- NormalError
#   |     +- XPathElementNotFoundError
#   |     +- UiObjectNotFoundError
#   |     +- AppNotFoundError
#   |     +- SessionBrokenError  
#   +- DeviceError
#      +- InputIMEError
#      +- HTTPError
#      +- ConnectError
#      +- AdbShellError
#      +- AdbBroadcastError
#      +- APKSignatureError
#      +- UiAutomationError
#         +- UiAutomationNotConnectedError
#         +- InjectPermissionError
#         +- LaunchUiAutomationError
#         +- AccessibilityServiceAlreadyRegisteredError


class BaseException(Exception):
    """ base error for uiautomator2 """

## DeviceError
class DeviceError(BaseException): ...
class AdbShellError(DeviceError):...
class ConnectError(DeviceError):...
class HTTPError(DeviceError):...
class HTTPTimeoutError(HTTPError):...
class AdbBroadcastError(DeviceError):...

class UiAutomationError(DeviceError):...
class InputIMEError(DeviceError):...

class UiAutomationNotConnectedError(UiAutomationError):...    
class InjectPermissionError(UiAutomationError):... #开发者选项中: 模拟点击没有打开
class APKSignatureError(UiAutomationError):...
class LaunchUiAutomationError(UiAutomationError):...
class AccessibilityServiceAlreadyRegisteredError(UiAutomationError):...


## RPCError
class RPCError(BaseException):
    pass

class RPCUnknownError(RPCError):...
class RPCInvalidError(RPCError):...
class HierarchyEmptyError(RPCError):...
class RPCStackOverflowError(RPCError):...


class NormalError(RPCError):
    pass

class XPathElementNotFoundError(NormalError):...
class SessionBrokenError(NormalError):... #only happens when app quit or crash
class UiObjectNotFoundError(NormalError):...
class AppNotFoundError(NormalError):...


# coding: utf-8
#

import contextlib
import functools
import inspect
import pathlib
import shlex
import sys
import threading
import typing
import warnings
from typing import Iterable, List, Tuple, Union

from PIL import Image

from uiautomator2._proto import Direction
from uiautomator2.exceptions import SessionBrokenError, UiObjectNotFoundError


@contextlib.contextmanager
def with_package_resource(filename: str) -> typing.Generator[pathlib.Path, None, None]:
    """
    Context manager to access a package asset file using importlib.resources.
    
    Args:
        filename (str): The name of the file to locate.
    
    Yields:
        str: The full path to the located file.
    """
    try:
        from importlib.resources import as_file, files
    except ImportError:
        # For Python < 3.9
        from importlib_resources import as_file, files
    anchor = files("uiautomator2") / filename
    with as_file(anchor) as f:
        if f.exists():
            yield f
            return
    
    # check if binary folder contains
    binary_path = pathlib.Path(sys.argv[0])
    if (binary_path / filename).exists():
        yield binary_path / filename
        return
    
    # check if running program directory contains
    if (pathlib.Path.cwd() / filename).exists():
        yield pathlib.Path.cwd() / filename
        return
    
    raise FileNotFoundError(f"Resource {filename} not found in uiautomator2 package.")
    

def check_alive(fn):
    @functools.wraps(fn)
    def inner(self, *args, **kwargs):
        if not self.running():
            raise SessionBrokenError(self._pkg_name)
        return fn(self, *args, **kwargs)

    return inner


_cached_values = {}


def cache_return(fn):
    @functools.wraps(fn)
    def inner(*args, **kwargs):
        key = (fn, args, frozenset(kwargs.items()))
        value = _cached_values.get(key)
        if value is not None:
            return value

        _cached_values[key] = ret = fn(*args, **kwargs)
        return ret

    return inner


def hooks_wrap(fn):
    @functools.wraps(fn)
    def inner(self, *args, **kwargs):
        name = fn.__name__.lstrip('_')
        self.server.hooks_apply("before", name, args, kwargs, None)
        ret = fn(self, *args, **kwargs)
        self.server.hooks_apply("after", name, args, kwargs, ret)

    return inner


# Will be removed in the future
def wrap_wait_exists(fn):
    @functools.wraps(fn)
    def inner(self, *args, **kwargs):
        timeout = kwargs.pop('timeout', self.wait_timeout)
        if not self.wait(timeout=timeout):
            raise UiObjectNotFoundError({
                'code': -32002,
                'message': self.selector.__str__()
            })
        return fn(self, *args, **kwargs)

    return inner


def intersect(rect1, rect2):
    top = rect1["top"] if rect1["top"] > rect2["top"] else rect2["top"]
    bottom = rect1["bottom"] if rect1["bottom"] < rect2["bottom"] else rect2[
        "bottom"]
    left = rect1["left"] if rect1["left"] > rect2["left"] else rect2["left"]
    right = rect1["right"] if rect1["right"] < rect2["right"] else rect2[
        "right"]
    return left, top, right, bottom


class Exists(object):
    """Exists object with magic methods."""
    def __init__(self, uiobject):
        self.uiobject = uiobject

    def __nonzero__(self):
        """Magic method for bool(self) python2 """
        return self.uiobject.jsonrpc.exist(self.uiobject.selector)

    def __bool__(self):
        """ Magic method for bool(self) python3 """
        return self.__nonzero__()

    def __call__(self, timeout=0):
        """Magic method for self(args).

        Args:
            timeout (float): exists in seconds
        """
        if timeout:
            return self.uiobject.wait(timeout=timeout)
        return bool(self)

    def __repr__(self):
        return str(bool(self))


def list2cmdline(args: Union[str, list, tuple]) -> str:
    if isinstance(args, str):
        return args
    return ' '.join(list(map(shlex.quote, args)))


def inject_call(fn, *args, **kwargs):
    """
    Call function without known all the arguments

    Args:
        fn: function
        args: arguments
        kwargs: key-values
    
    Returns:
        as the fn returns
    """
    assert callable(fn), "first argument must be callable"

    st = inspect.signature(fn)
    fn_kwargs = {
        key: kwargs[key]
        for key in st.parameters.keys() if key in kwargs
    }
    ba = st.bind(*args, **fn_kwargs)
    ba.apply_defaults()
    return fn(*ba.args, **ba.kwargs)


def natualsize(size: int) -> str:
    _KB = 1 << 10
    _MB = 1 << 20
    _GB = 1 << 30

    if size >= _GB:
        return '{:.1f} GB'.format(size / _GB)
    elif size >= _MB:
        return '{:.1f} MB'.format(size / _MB)
    else:
        return '{:.1f} KB'.format(size / _KB)


def swipe_in_bounds(d: "uiautomator2.Device",
                    bounds: Tuple[int, int, int, int],
                    direction: Union[Direction, str],
                    scale: float = 0.6):
    """
    Args:
        d: Device object
        bounds: list of [lx, ly, rx, ry]
        direction: one of ["left", "right", "up", "down"]
        scale: percent of swipe, range (0, 1.0)
    
    Raises:
        AssertionError, ValueError
    """
    def _swipe(_from, _to):
        d.swipe(_from[0], _from[1], _to[0], _to[1])

    assert 0 < scale <= 1.0
    assert len(bounds) == 4

    lx, ly, rx, ry = bounds
    width, height = rx - lx, ry - ly

    h_offset = int(width * (1 - scale)) // 2
    v_offset = int(height * (1 - scale)) // 2

    left = lx + h_offset, ly + height // 2
    up = lx + width // 2, ly + v_offset
    right = rx - h_offset, ly + height // 2
    bottom = lx + width // 2, ry - v_offset

    if direction == Direction.LEFT:
        _swipe(right, left)
    elif direction == Direction.RIGHT:
        _swipe(left, right)
    elif direction == Direction.UP:
        _swipe(bottom, up)
    elif direction == Direction.DOWN:
        _swipe(up, bottom)
    else:
        raise ValueError("Unknown direction:", direction)


def thread_safe_wrapper(fn: typing.Callable):
    @functools.wraps(fn)
    def inner(self, *args, **kwargs):
        if not hasattr(self, "_lock"):
            self._lock = threading.Lock()

        with self._lock:
            return fn(self, *args, **kwargs)
    
    return inner



def is_version_compatiable(expect_version: str, actual_version: str) -> bool:
    """
    Check if the actual version is compatiable with the expect version

    Args:
        expect_version: expect version, e.g. 1.0.0
        actual_version: actual version, e.g. 1.0.0

    Returns:
        bool: True if compatiable, otherwise False
    """
    def _parse_version(version: str):
        return tuple(map(int, version.split(".")))

    evs = _parse_version(expect_version)
    avs = _parse_version(actual_version)
    assert len(evs) == len(avs) == 3, "version format error"
    if evs[0] == avs[0]:
        if evs[1] < avs[1]:
            return True
        if evs[1] == avs[1]:
            return evs[2] <= avs[2]
    return False


def deprecated(reason):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            warnings.warn(f"Function '{func.__name__}' is deprecated: {reason}", DeprecationWarning, stacklevel=2)
            return func(*args, **kwargs)
        return wrapper
    return decorator


def image_convert(im: Image.Image, format: str):
    if format == "pillow":
        return im
    if format == "opencv":
        try:
            import cv2
            import numpy as np
            im = im.convert("RGB")
            return cv2.cvtColor(np.array(im), cv2.COLOR_RGB2BGR)
        except ImportError:
            warnings.warn("missing lib: cv2 or numpy")
            raise
    raise ValueError("Unsupported format:", format)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Thu Apr 25 2024 15:08:43 by codeskyblue
"""

import abc
from typing import Any, List, NamedTuple, Tuple, Union
import typing
import adbutils
from PIL import Image
from uiautomator2._proto import Direction


class ShellResponse(NamedTuple):
    output: str
    exit_code: int
    
    

class AbstractUiautomatorServer(abc.ABC):
    @abc.abstractmethod
    def start_uiautomator(self):
        pass

    @abc.abstractmethod
    def stop_uiautomator(self):
        pass

    @abc.abstractmethod
    def jsonrpc_call(self, method: str, params: Any = None) -> Any:
        pass



class AbstractShell(abc.ABC):
    @abc.abstractmethod
    def shell(self, cmdargs: Union[List[str], str]) -> ShellResponse:
        pass

    @property
    @abc.abstractmethod
    def adb_device(self) -> adbutils.AdbDevice:
        pass
    
    @property
    @abc.abstractmethod
    def jsonrpc(self) -> typing.Any:
        pass
    

class AbstractXPathBasedDevice(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def click(self, x: int, y: int):
        pass
    
    @abc.abstractmethod
    def long_click(self, x: int, y: int):
        pass

    @abc.abstractmethod
    def send_keys(self, text: str):
        pass

    @abc.abstractmethod
    def swipe(self, fx: int, fy: int, tx: int, ty: int, duration: float):
        """ duration is float type, indicate seconds """
    
    @abc.abstractmethod
    def swipe_ext(self, direction: Direction, scale: float):
        pass
    
    @abc.abstractmethod
    def window_size(self) -> Tuple[int, int]:
        """ return (width, height) """
    
    @abc.abstractmethod
    def dump_hierarchy(self) -> str:
        """ return xml content """
    
    @abc.abstractmethod
    def screenshot(self) -> Image.Image:
        """ return PIL.Image.Image """



import enum

SCROLL_STEPS = 55
HTTP_TIMEOUT = 300

class Direction(str, enum.Enum):
    LEFT = "left"
    RIGHT = "right"
    UP = "up"
    DOWN = "down"

    # 垂直操作
    FORWARD = "up"
    BACKWARD = "down"

    # 水平操作
    HORIZ_FORWARD = "left"
    HORIZ_BACKWARD = "right"


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Wed May 22 2024 16:23:56 by codeskyblue
"""

import base64
from dataclasses import dataclass
import logging
from pathlib import Path
import re
import time
from typing import Dict, List, Optional, Union
import warnings

import adbutils
from retry import retry

from uiautomator2.abstract import AbstractShell
from uiautomator2.exceptions import AdbBroadcastError, DeviceError, InputIMEError
from uiautomator2.utils import deprecated, with_package_resource


logger = logging.getLogger(__name__)

@dataclass
class BroadcastResult:
    code: Optional[int]
    data: Optional[str]


BORADCAST_RESULT_OK = -1
BROADCAST_RESULT_CANCELED = 0



class InputMethodMixIn(AbstractShell):
    # @property
    # def clipboard(self) -> Optional[str]:
    #     result = self._broadcast("ADB_KEYBOARD_GET_CLIPBOARD")
    #     if result.code == BORADCAST_RESULT_OK:
    #         return base64.b64decode(result.data).decode('utf-8')
    #     # jsonrpc.getClipboard is not OK for now
    #     return None
    
    @property
    def __ime_id(self) -> str:
        return 'com.github.uiautomator/.AdbKeyboard'

    def set_input_ime(self, enable: bool = True):
        """ Enable of Disable InputIME """
        if not enable:
            self.shell(['ime', 'disable', self.__ime_id])
            return
        if self.current_ime() == self.__ime_id:
            return
        # prepare ime
        if self.__ime_id not in self.__get_ime_list():
            self._setup_ime()
        assert self.__ime_id in self.__get_ime_list()
        
        self.shell(['ime', 'enable', self.__ime_id])
        self.shell(['ime', 'set', self.__ime_id])
        self.shell(['settings', 'put', 'secure', 'default_input_method', self.__ime_id])
        self._wait_ime_ready()
    
    def is_input_ime_installed(self) -> bool:
        return self.__ime_id in self.__get_ime_list()
        
    def _setup_ime(self):
        logger.debug("installing AdbKeyboard ime")
        with with_package_resource("assets/app-uiautomator.apk") as ime_apk_path:
            try:
                self.adb_device.install(str(ime_apk_path), nolaunch=True, uninstall=True)
            except adbutils.AdbError as e:
                self.adb_device.uninstall(self.__ime_id.split('/')[0])
                self.adb_device.install(str(ime_apk_path), nolaunch=True, uninstall=True)
            
        # wait for ime registered
        for _ in range(10):
            if self.__ime_id in self.__get_ime_list():
                return
            time.sleep(.3)
        raise InputIMEError("install AdbKeyboard ime failed")
    
    def _broadcast(self, action: str, extras: Dict[str, str] = {}) -> BroadcastResult:
        # requires ATX 2.4.0+
        args = ['am', 'broadcast', '-a', action]
        for k, v in extras.items():
            if isinstance(v, int):
                args.extend(['--ei', k, str(v)])
            else:
                args.extend(['--es', k, v])
        # Example output: result=-1 data="success"
        output = self.shell(args).output
        m_result = re.search(r'result=(-?\d+)', output)
        m_data = re.search(r'data="([^"]+)"', output)
        result = int(m_result.group(1)) if m_result else None
        data = m_data.group(1) if m_data else None
        return BroadcastResult(result, data)
    
    @retry(AdbBroadcastError, tries=3, delay=1, jitter=0.5)
    def _must_broadcast(self, action: str, extras: Dict[str, str] = {}):
        result = self._broadcast(action, extras)
        if result.code != BORADCAST_RESULT_OK:
            raise AdbBroadcastError(f"broadcast {action} failed: {result.data}")

    def send_keys(self, text: str):
        try:
            self.set_input_ime()
            btext = text.encode('utf-8')
            base64text = base64.b64encode(btext).decode()
            cmd = "ADB_KEYBOARD_INPUT_TEXT"
            self._must_broadcast(cmd, {"text": base64text})
            return True
        except AdbBroadcastError:
            warnings.warn(
                "set FastInputIME failed. use \"d(focused=True).set_text instead\"",
                Warning)
            return self(focused=True).set_text(text)
        
    def send_action(self, code: Union[str, int] = None):
        """
        Simulate input method edito code

        Args:
            code (str or int): input method editor code

        Examples:
            send_action("search"), send_action(3)

        Refs:
            https://developer.android.com/reference/android/view/inputmethod/EditorInfo
        """
        self.set_input_ime(True)
        __alias = {
            "go": 2,
            "search": 3,
            "send": 4,
            "next": 5,
            "done": 6,
            "previous": 7,
        }
        if isinstance(code, str):
            code = __alias.get(code, code)
        if code:
            self._must_broadcast('ADB_KEYBOARD_EDITOR_CODE', {"code": str(code)})
        else:
            self._must_broadcast('ADB_KEYBOARD_SMART_ENTER')

    def clear_text(self):
        self.set_input_ime(True)
        self._must_broadcast('ADB_KEYBOARD_CLEAR_TEXT')

    def current_ime(self) -> str:
        """ Current input method
        Returns:
            ime_method

        Example output:
            "com.github.uiautomator/.FastInputIME"
        """
        return self.shell(['settings', 'get', 'secure', 'default_input_method']).output.strip()
        # _INPUT_METHOD_RE = re.compile(r'mCurMethodId=([-_./\w]+)')
        # dim, _ = self.shell(['dumpsys', 'input_method'])
        # m = _INPUT_METHOD_RE.search(dim)
        # method_id = None if not m else m.group(1)
        # shown = "mInputShown=true" in dim
        # return (method_id, shown)
    
    def _wait_ime_ready(self, timeout: float = 5.0) -> bool:
        """ Wait for input method is ready """
        deadline = time.time() + timeout
        while time.time() < deadline:
            if self.current_ime() == self.__ime_id:
                return True
            time.sleep(0.1)
        return False
    
    def __get_ime_list(self) -> List[str]:
        ret = self.shell(['ime', 'list', '-s', '-a'])
        return ret.output.strip().splitlines(keepends=False)

    def hide_keyboard(self):
        """ Hide keyboard """
        self.set_input_ime()
        self._must_broadcast('ADB_KEYBOARD_HIDE')
        
    @deprecated(reason="use set_input_ime instead")
    def set_fastinput_ime(self, enable: bool = True):
        return self.set_input_ime(enable)
    
    @deprecated(reason="use set_input_ime instead")
    def wait_fastinput_ime(self, timeout=5.0):
        """ wait FastInputIME is ready (Depreacated in version 3.1) """
        pass
